#if !defined( PYCXX_STANDARD_EXCEPTION )
#pragma error( "define PYCXX_STANDARD_EXCEPTION before including" )
#endif

PYCXX_STANDARD_EXCEPTION( SystemExit,       BaseException )
PYCXX_STANDARD_EXCEPTION( KeyboardInterrupt,BaseException )
PYCXX_STANDARD_EXCEPTION( GeneratorExit,    BaseException )
#if !defined( PYCXX_6_2_COMPATIBILITY )
PYCXX_STANDARD_EXCEPTION( Exception,        BaseException )
#endif
PYCXX_STANDARD_EXCEPTION(     StopIteration,    Exception )
PYCXX_STANDARD_EXCEPTION(     StandardError,    Exception )
PYCXX_STANDARD_EXCEPTION(         BufferError,      StandardError )
PYCXX_STANDARD_EXCEPTION(         ArithmeticError,  StandardError )
PYCXX_STANDARD_EXCEPTION(             FloatingPointError,   ArithmeticError )
PYCXX_STANDARD_EXCEPTION(             OverflowError,        ArithmeticError )
PYCXX_STANDARD_EXCEPTION(             ZeroDivisionError,    ArithmeticError )
PYCXX_STANDARD_EXCEPTION(         AssertionError,   StandardError )
PYCXX_STANDARD_EXCEPTION(         AttributeError,   StandardError )
PYCXX_STANDARD_EXCEPTION(         EnvironmentError, StandardError )
PYCXX_STANDARD_EXCEPTION(             IOError,          EnvironmentError )
PYCXX_STANDARD_EXCEPTION(             OSError,          EnvironmentError )
#ifdef MS_WINDOWS
PYCXX_STANDARD_EXCEPTION(                 WindowsError, OSError )
#endif
PYCXX_STANDARD_EXCEPTION(         EOFError,         StandardError )
PYCXX_STANDARD_EXCEPTION(         ImportError,      StandardError )
PYCXX_STANDARD_EXCEPTION(         LookupError,      StandardError )
PYCXX_STANDARD_EXCEPTION(             IndexError,       LookupError )
PYCXX_STANDARD_EXCEPTION(             KeyError,         LookupError )
PYCXX_STANDARD_EXCEPTION(         MemoryError,      StandardError )
PYCXX_STANDARD_EXCEPTION(         NameError,        StandardError )
PYCXX_STANDARD_EXCEPTION(             UnboundLocalError,NameError )
PYCXX_STANDARD_EXCEPTION(         ReferenceError,   StandardError )
PYCXX_STANDARD_EXCEPTION(         RuntimeError,     StandardError )
PYCXX_STANDARD_EXCEPTION(             NotImplementedError,  RuntimeError )
PYCXX_STANDARD_EXCEPTION(         SyntaxError,      StandardError )
PYCXX_STANDARD_EXCEPTION(             IndentationError,     SyntaxError )
PYCXX_STANDARD_EXCEPTION(                 TabError,             IndentationError )
PYCXX_STANDARD_EXCEPTION(         SystemError,      StandardError )
PYCXX_STANDARD_EXCEPTION(         TypeError,        StandardError )
PYCXX_STANDARD_EXCEPTION(         ValueError,       StandardError )
PYCXX_STANDARD_EXCEPTION(             UnicodeError,     ValueError )
PYCXX_STANDARD_EXCEPTION(                 UnicodeDecodeError,   UnicodeError )
PYCXX_STANDARD_EXCEPTION(                 UnicodeEncodeError,   UnicodeError )
PYCXX_STANDARD_EXCEPTION(                 UnicodeTranslateError,UnicodeError )
